# -*- coding: utf-8 -*-
"""
Single stub matching analysis
ABCD method and new method

Timothy D. Drysdale
tim.drysdale@open.ac.uk

History
20 Feb 2017 - updated to latest example for the paper
31 May 2016 - added open circuit stub
8 Sep 2015 - created
"""

import numpy as np
import matplotlib.pyplot as plt 
from matplotlib import rcParams
rcParams['text.usetex'] = True


def coax_properties(d_inner, d_outer, eps_r, mu_r=1):
    Z0 = 138.0 * np.log10(d_outer / d_inner * eps_r**-0.5) 
    return(Z0) 

def pack(A,B,C,D):
    return(np.array(((A,B),(C,D)), dtype='complex'))
    
def unpack(ABCD):
    
    if np.size(np.shape(ABCD)) == 2:
        A = ABCD[0,0]
        B = ABCD[0,1]
        C = ABCD[1,0]
        D = ABCD[1,1]
    else:    
        A = ABCD[0,0,:]
        B = ABCD[0,1,:]
        C = ABCD[1,0,:]
        D = ABCD[1,1,:]
    
    return(A, B, C, D)
    
def get_ABCD_tline(length,beta,Z0):
    bl = beta * length
    
    A = np.cos(bl)
    B = 1j * Z0 * np.sin(bl)
    C = 1j / Z0 * np.sin(bl)
    D = np.cos(bl)     

    return(pack(A, B, C, D))
    
def get_ABCD_stub(length,beta,Z0,short=True):
    
    if short==True: 
        #short circuit
        Zin = 1.0j * Z0 * np.tan(beta *length)
    else: 
        #open circuit
        Zin = -1.0j * Z0 / np.tan(beta * length) 

    A = 1.0
    B = 0.0    
    C = 1.0 / Zin 
    D = 1.0

    return(pack(A, B, C, D))
    
def invert_ABCD(ABCD):
    if np.size(np.shape(ABCD)) == 2:
        return(np.linalg.inv(np.matrix(ABCD)))
    else:
        invABCD = {}
        for m in np.arange(np.size(ABCD,2)):
            temp_result = np.linalg.inv(np.matrix(ABCD[:,:,m]))
            invABCD = np.append(ABCD, np.array(temp_result,dtype='complex'))
 
        invABCD = np.reshape(invABCD[1:],(2,2,-1))        
        repacked_inv  = pack(invABCD[0,0,:], invABCD[0,1,:],
                             invABCD[1,0,:], invABCD[1,1,:])
        return(repacked_inv)     


def get_ABCD_load(beta,ZL):
    A = 1.0 * np.ones_like(beta)
    B = 0.0 * np.ones_like(beta)
    C = 1.0 / ZL * np.ones_like(beta)
    D = 1.0 * np.ones_like(beta)
    return(pack(A, B, C, D))
    
def multiply(left,right):
    if np.size(np.shape(left)) == 2:
        return(np.matrix(left) * np.matrix(right))
    else:
        ABCD = {}
        for m in np.arange(np.size(left,2)):
            temp_result = np.matrix(left[:,:,m]) * np.matrix(right[:,:,m])
            ABCD = np.append(ABCD, np.array(temp_result,dtype='complex'))

        ABCD = np.reshape(ABCD[1:],(2,2,-1))        
        
        repacked  = pack(ABCD[0,0,:], ABCD[0,1,:],
                         ABCD[1,0,:], ABCD[1,1,:])
        return(ABCD)

def calculate_Zin(system_ABCD,ZL):
    (A,B,C,D) = unpack(system_ABCD)
    Zin = (A * ZL + B) / (C * ZL + D)
    return(Zin)
    
def calculate_Zx(beta,L,x,Z0,ZL):
    gamma = 1j * beta 
    E = 1
    top_line = ZL + Z0*np.tanh(gamma*(L-x))
    bot_line = Z0 + ZL*np.tanh(gamma*(L-x))
    Zx = top_line / bot_line * Z0 
    
    Ix = E / (Zx*np.cosh(gamma*x) + Z0*np.sinh(gamma*x))
    Vx = Zx*Ix
    return(Vx,Ix,Zx)

def calculate_Vfb(beta,x,Vs,Z0,Zin):
    Vf = 0.5 * (1 + Z0 / Zin) * np.exp(-1j * beta * x)    
    Vb = 0.5 * (1 - Z0 / Zin) * np.exp(1j * beta * x)
    return(Vf,Vb)
    
   
def calculate_Vx(beta,x,Vs,Z0,Zin):
    Vx = Vs * (np.cos(beta * x) - 1j * Z0 / Zin * np.sin(beta * x))     
    return(Vx)
    
def calculate_Ix(beta,x,Vs,Z0,Zin):
    Ix = Vs * (np.cos(beta * x) / Zin - 1j / Z0 * np.sin(beta * x))     
    return(Ix)   

  
def one_stub(vf = 0.67, freq = 120e6, Z0 = 75, ZL = 44.8-107j,
             L_in = 10e-2, L_load = 57.7e-2, L_stub = 54.7e-2, short = True,
             filestub = 'onestub'):

    v = 3e8 * vf 
       
    wavelength = v / freq    
    beta = 2.0 * np.pi / wavelength 
    
    lengths = np.linspace(0,L_in,100)
    load_lengths = np.linspace(0,L_load,100)
    stub_lengths = np.linspace(0,L_stub,100)
    
    Zx_load1 = [] 
    Zx_load2 = [] 
    Zx_load3 = []

    Vx_load1 = [] 
    Vx_load2 = []
    Vx_load3 = []
    
    Vxx_load1 = [] 
    Vxx_load2 = [] 
    Vxx_load3 = []

    Ixx_load1 = [] 
    Ixx_load2 = []
    Ixx_load3 = []
    
    Ix_load1 = [] 
    Ix_load2 = [] 
    Ix_load3 = []
    
    B = beta
    L = L_in
    
    Vs = 1
        
    '''
    look down line from input
    '''
    
    line_in = get_ABCD_tline(L_in,B,Z0)  
    stub = get_ABCD_stub(L_stub,B,Z0,short)
    line_load = get_ABCD_tline(L_load,B,Z0)  
    
    system = multiply(line_in,stub)    
    system = multiply(system,line_load)
    Zin = calculate_Zin(system,ZL)      
    
    for x in lengths:    
        (Vx,Ix,Zx) = calculate_Zx(B,L,x,Z0,ZL)   #this ZL needs updateing to Zin for rest of system
        Zx_load1 = np.append(Zx_load1,Zx)    
        Vx_load1 = np.append(Vx_load1,Vx) 
        Ix_load1 = np.append(Ix_load1,Ix) 
        
        Vxx = calculate_Vx(B,x,Vs,Z0,Zin)
        Vxx_load1 = np.append(Vxx_load1,Vxx) 

        Ixx = calculate_Ix(B,x,Vs,Z0,Zin)
        Ixx_load1 = np.append(Ixx_load1,Ixx)    
   
    '''
    From stub junction, look down line to load
    '''
    
    Vs = Vxx
    Zin = calculate_Zin(line_load,ZL)      
    
    for x in load_lengths:    
        (Vx,Ix,Zx) = calculate_Zx(B,L,x,Z0,ZL)
        Zx_load2 = np.append(Zx_load2,Zx)    
        Vx_load2 = np.append(Vx_load2,Vx) 
        Ix_load2 = np.append(Ix_load2,Ix) 
       
        Vxx = calculate_Vx(B,x,Vs,Z0,Zin)
        Vxx_load2 = np.append(Vxx_load2,Vxx) 

        Ixx = calculate_Ix(B,x,Vs,Z0,Zin)
        Ixx_load2 = np.append(Ixx_load2,Ixx)        

    '''
    From stub junction, look down line to stub
    '''
    
    if short == True:
        Zshort = 0
    else:
        Zshort = 9e99
        
    line_stub = get_ABCD_tline(L_stub,B,Z0)
    Zin = calculate_Zin(line_stub,Zshort)    
    
    for x in stub_lengths:    
        (Vx,Ix,Zx) = calculate_Zx(beta,L,x,Z0,0)
        Zx_load3 = np.append(Zx_load3,Zx)    
        Vx_load3 = np.append(Vx_load3,Vx) 
        Ix_load3 = np.append(Ix_load3,Ix) 
       
        Vxx = calculate_Vx(beta,x,Vs,Z0,Zin)
        Vxx_load3 = np.append(Vxx_load3,Vxx) 

        Ixx = calculate_Ix(beta,x,Vs,Z0,Zin)
        Ixx_load3 = np.append(Ixx_load3,Ixx)  
        

    Vnorm = np.max(np.abs(Vxx_load1))     
    Inorm = np.max(np.abs(Ixx_load1))
    
    VI = np.abs(Vxx_load1) / Vnorm
    VL = np.abs(Vxx_load2) / Vnorm
    VS = np.abs(Vxx_load3) / Vnorm
    
    II = np.abs(Ixx_load1) / Inorm
    IL = np.abs(Ixx_load2) / Inorm
    IS = np.abs(Ixx_load3) / Inorm

    LI = lengths/1e-3
    LL = load_lengths/1e-3
    LS = stub_lengths/1e-3
    
    return(VI, VL, VS, II, IL, IS, LI, LL, LS)    

'''
New analysis method
'''    
    
def calc_beta(freq, vf):
    velocity = 3e8 * vf
    wavelength = velocity /freq
    beta = 2.0 * np.pi / wavelength 
    return(beta)
    
    
def calc_rhoL0(L_load, beta, Z0, ZL):
    rhoL = (ZL-Z0)/(ZL+Z0)
    rhoL0 = rhoL * np.exp(-2j * beta * L_load)
    return(rhoL0)    


def calc_rhoS0(L_stub, beta, Z0, short = True):
    if short == True:
        rhoS = -1 
    else: 
        rhoS = +1 

    rhoS0 = rhoS * np.exp(-2j * beta * L_stub)
    return(rhoS0)


def calc_rhoLS(rhoL0, rhoS0):
    topline = rhoL0 + rhoS0 + (3 * rhoL0 * rhoS0) - 1
    botline = rhoL0 + rhoS0 - (rhoL0 * rhoS0) + 3
    return(topline / botline)
 
 
def calc_VS_factor(rhoL0, rhoS0, rhoLS):
    topline = (1 + rhoL0) * (1 - rhoLS) 
    botline = 2 * (1 - (rhoL0 * rhoS0))
    return(topline /botline) 
 
def calc_VL_factor(rhoL0, rhoS0, rhoLS):
    topline = (1 + rhoS0) * (1 - rhoLS) 
    botline = 2 * (1 - (rhoL0 * rhoS0))
    return(topline / botline)   

def calc_VL(VLfactor, rhoL0, beta, x, VIp):
    V = VIp * VLfactor*(np.exp(-1j * beta * x) + (rhoL0 * np.exp(+1j * beta *x)))    
    return(V)

def calc_VS(VSfactor, rhoS0, beta, x, VIp):
    V = VIp * VSfactor*(np.exp(-j1 * beta * x) + (rhoS0 * np.exp(+1j * beta *x)))    
    return(V)
    
    
def calc_VI(rhoLS, beta, x, VIp):
    V = VIp * (np.exp(+1j * beta * x) + (rhoLS * np.exp(-1j * beta *x)))
    return(V)
        

def one_stub_Alan(vf = 0, freq = 0, Z0 = 0, ZL = 0,
             L_in = 0, L_load = 0, L_stub = 0, short = True,
             filestub = 'Alanonestub'):
                 
    beta = calc_beta(freq, vf)
    lengths = np.linspace(-L_in,0,100) 
    load_lengths = np.linspace(0,L_load,100)
    stub_lengths = np.linspace(0,L_stub,100)

    VIp = 1 #source voltage
    
    #use p to distinguish from function's rho
    pS0 = calc_rhoS0(L_stub, beta, Z0, short) 
    pL0 = calc_rhoL0(L_load, beta, Z0, ZL)
    pLS = calc_rhoLS(pL0, pS0)
    
    VSfactor = calc_VS_factor(pL0, pS0, pLS)
    VLfactor = calc_VL_factor(pL0, pS0, pLS)
   
    VSraw = calc_VL(VSfactor, pS0, beta, stub_lengths, VIp)
    VLraw = calc_VL(VLfactor, pL0, beta, load_lengths, VIp)
    VIraw = calc_VI(pLS, beta, lengths, VIp)
    
    IIraw = VIraw / Z0
    ILraw = VLraw / Z0
    ISraw = VSraw / Z0
    
    Vnorm = np.max(np.abs(VIraw))     
    Inorm = np.max(np.abs(IIraw))
    
    VI = np.abs(VIraw) / Vnorm
    VL = np.abs(VLraw) / Vnorm
    VS = np.abs(VSraw) / Vnorm
    
    II = np.abs(IIraw) / Inorm  
    IL = np.abs(ILraw) / Inorm
    IS = np.abs(ISraw) / Inorm

    LI = -1 * lengths/1e-3
    LL = load_lengths/1e-3
    LS = stub_lengths/1e-3    
    
    return(VI, VL, VS, II, IL, IS, LI, LL, LS)        


def plot_comparison(vf = 0,
             freq = 0, Z0 = 0, ZL = 0, 
             L_in = 0, L_load = 0, L_stub = 0, short = True,
             filestub = 'default', legend_loc = ''):
    
    #ABCD method             
    (VI, VL, VS, II, IL, IS, LI, LL, LS) = one_stub(vf = vf,
             freq = freq, Z0 = Z0, ZL = ZL, 
             L_in = L_in, L_load = L_load, L_stub = L_stub, short =short,
             filestub = filestub) 
             
        
    fig, ax = plt.subplots(1,3,figsize=(12,4),sharey=True) 
    
    #need these commas - single element tuple unpacking
    lv1, = ax[0].plot(LI,VI,'k-')    
    lv2, = ax[1].plot(LL,VL,'k-')    
    lv3, = ax[2].plot(LS,VS,'k-')  
    
    
    #New analysis
    (VIa, VLa, VSa, IIa, ILa, ISa, LIa, LLa, LSa)  = one_stub_Alan(vf = vf,
             freq = freq, Z0 = Z0, ZL = ZL, 
             L_in = L_in, L_load = L_load, L_stub = L_stub,short = short,
             filestub = filestub) 
             
    lv1a, = ax[0].plot(LIa,VIa, 's', markevery = 9, markerfacecolor='none')    
    lv2a, = ax[1].plot(LLa,VLa, 's', markevery = 9, markerfacecolor='none')    
    lv3a, = ax[2].plot(LSa,VSa, 's', markevery = 9, markerfacecolor='none')     
    
   
    ax[0].set_title('Line from input') 
    ax[1].set_title('Line to load')
    ax[2].set_title('Stub')
    
    if legend_loc == '':
       legend_loc = 'upper right'
    
    plt.legend((lv1, lv1a), ("V$_{x}$ ABCD", 
               'V$_{x}$ Analytical'), loc=legend_loc, ncol = 1)
                                   
    plt.suptitle('Signals on stub matched network')
    
    ax[0].set_xlabel('distance / mm')    
    ax[1].set_xlabel('distance / mm')     
    ax[2].set_xlabel('distance / mm')    
    max_y = np.max([VI,VL,VS])    
    ax[0].set_ylim([0,max_y*1.05])

    plt.savefig('%s-total_voltages_Z0=%0.0f_ZL=%0.0f+%0.0fj.png'%(filestub,np.abs(Z0),np.real(ZL), np.imag(ZL)),dpi=300)
    plt.show()                 

    return()
                                               
if __name__ == "__main__":

    freq = 610e6
    L_in = 0.2 
    L_load = 0.215*0.492
    L_stub = 0.303*0.492  
    Z0 = 50
    ZL = 300+250j 
    filestub = 'open_2017'
    vf = 1.0
    short = False
    
    plot_comparison(vf = vf,
             freq = freq, Z0 = Z0, ZL = ZL, 
             L_in = L_in, L_load = L_load, L_stub = L_stub, short =short,
             filestub = filestub, legend_loc='lower right') 
    
    

      
             

